import os
import json
import pytest
from unittest.mock import ANY

from memorious.core import tags
from memorious.operations.fetch import fetch, session
from memorious.operations.parse import parse
from memorious.operations.initializers import seed, sequence, dates, enumerate
from memorious.operations.store import directory


@pytest.mark.parametrize("url,call_count", [
    ("https://httpbin.org/html", 1),
    ("https://occrp.org/", 0),
    ("https://httpbin.org/status/418", 0)
])
def test_fetch(url, call_count, context, mocker):
    rules = {'pattern': 'https://httpbin.org/*'}
    context.params["rules"] = rules
    mocker.patch.object(context, "emit")
    fetch(context, {"url": url})
    assert context.emit.call_count == call_count


def test_session(context, mocker):
    context.params["user"] = "user"
    context.params["password"] = "password"
    context.params["user_agent"] = "Godzilla Firehose 0.1"
    context.params["url"] = "https://httpbin.org/get"
    data = {"hello": "world"}
    mocker.patch.object(context.http, "save")
    mocker.patch.object(context, "emit")

    session(context, data)

    assert context.http.save.called_one_with()
    assert context.emit.called_one_with(data=data)
    assert (
        context.http.session.headers["User-Agent"] == "Godzilla Firehose 0.1"
    )
    assert context.http.session.headers["Referer"] == "https://httpbin.org/get"
    assert context.http.session.auth == ("user", "password")


def test_parse(context, mocker):
    url = "http://example.org/"
    result = context.http.get(url)
    data = result.serialize()

    mocker.patch.object(context, "emit")

    rules = {'pattern': 'https://httpbin.org/*'}
    context.params["store"] = rules
    context.params["meta"] = {
        "title": ".//h1",
        "description": ".//p"
    }
    parse(context, data)
    assert context.emit.call_count == 1
    context.emit.assert_called_once_with(rule="fetch", data=ANY)

    # cleanup tags
    tags.delete()

    context.http.result = None
    context.params["store"] = None
    parse(context, data)
    assert data['url'] == 'https://www.iana.org/domains/example'
    assert data['title'] == 'Example Domain'
    assert data['description'].startswith('This domain is for')
    assert context.emit.call_count == 3, data


def test_seed(context, mocker):
    context.params["url"] = None
    context.params["urls"] = [
        "http://httpbin.org/status/%(status)s"
    ]
    mocker.patch.object(context, "emit")
    seed(context, data={"status": 404})
    assert context.emit.call_count == 1
    context.emit.assert_called_once_with(
        data={"url": "http://httpbin.org/status/404"}
    )


def test_sequence(context, mocker):
    mocker.patch.object(context, "emit")

    context.params["start"] = 2
    context.params["stop"] = 11
    context.params["step"] = 3
    sequence(context, data={})
    assert context.emit.call_count == 3

    context.params["start"] = 7
    context.params["stop"] = 1
    context.params["step"] = -3
    sequence(context, data={})
    assert context.emit.call_count == 5


def test_dates(context, mocker):
    mocker.patch.object(context, "emit")
    mocker.patch.object(context, "recurse")
    context.params["format"] = '%d-%m-%Y'
    context.params["days"] = 3
    context.params["begin"] = "10-12-2012"
    context.params["end"] = "20-12-2012"
    dates(context, data={})
    assert context.emit.call_count == 1
    context.emit.assert_called_once_with(data={
        "date": "20-12-2012",
        "date_iso": "2012-12-20T00:00:00"
    })
    assert context.recurse.call_count == 1
    context.recurse.assert_called_once_with(
        data={"current": "17-12-2012"}
    )


def test_enumerate(context, mocker):
    mocker.patch.object(context, "emit")
    context.params["items"] = [1, 2, 3]
    enumerate(context, data={})
    assert context.emit.call_count == 3
    # expected_calls = [
    #     mocker.call(data={"item": 1}),
    #     mocker.call(data={"item": 2}),
    #     mocker.call(data={"item": 3}),
    # ]
    # assert context.emit.mock_calls == expected_calls
    # Ideally this should work, but currently it doesn't; because we pass
    # the reference to data dict around and then mutate it.


def test_directory(context):
    # echo user-agent
    url = "https://httpbin.org/user-agent"
    result = context.http.get(url, headers={"User-Agent": "Memorious Test"})
    data = result.serialize()
    directory(context, data)

    content_hash = data.get('content_hash')
    file_path = os.path.realpath(__file__)
    store_dir = os.path.normpath(os.path.join(
        file_path, "../testdata/data/store/occrp_web_site"
    ))

    raw_file_path = os.path.join(store_dir, content_hash + ".data.json")
    meta_file_path = os.path.join(store_dir, content_hash + ".json")
    assert os.path.exists(meta_file_path)
    assert os.path.exists(raw_file_path)

    with open(meta_file_path, "rb") as fh:
        assert json.load(fh)['content_hash'] == data['content_hash']
    with open(raw_file_path, "rb") as fh:
        assert b'"user-agent": "Memorious Test"' in fh.read()



import logging
from servicelayer.worker import Worker
from servicelayer.cache import make_key

from memorious import settings
from memorious.logic.context import Context
from memorious.logic.stage import CrawlerStage
from memorious.core import manager, conn, get_rate_limit

log = logging.getLogger(__name__)


class MemoriousWorker(Worker):
    def boot(self):
        self.scheduler = get_rate_limit('scheduler',
                                        unit=60,
                                        interval=settings.SCHEDULER_INTERVAL,
                                        limit=1)
        self.hourly = get_rate_limit('hourly', unit=3600, interval=1, limit=1)

    def periodic(self):
        if self.hourly.check():
            self.hourly.update()
            log.info("Running hourly tasks...")
            for crawler in manager:
                if crawler.should_timeout:
                    crawler.timeout()
        if self.scheduler.check() and not settings.DEBUG:
            log.info("Running scheduled crawlers ...")
            self.scheduler.update()
            manager.run_scheduled()
        self.timeout_expiration_check()

    def handle(self, task):
        data = task.payload
        stage = CrawlerStage.detach_namespace(task.stage.stage)
        state = task.context
        context = Context.from_state(state, stage)
        context.execute(data)

    def after_task(self, task):
        if task.job.is_done():
            stage = CrawlerStage.detach_namespace(task.stage.stage)
            state = task.context
            context = Context.from_state(state, stage)
            context.crawler.aggregate(context)
        self.timeout_expiration_check()

    def get_stages(self):
        all_stages = set({stage.namespaced_name for _, stage in manager.stages})  # noqa
        stages_on_timeout_key = make_key('memorious', 'timeout_stages')
        stages_on_timeout = conn.smembers(stages_on_timeout_key)
        if stages_on_timeout:
            return list(all_stages - set(stages_on_timeout))
        return all_stages

    def timeout_expiration_check(self):
        stages_on_timeout_key = make_key('memorious', 'timeout_stages')
        stages_on_timeout = conn.smembers(stages_on_timeout_key)
        for stage in stages_on_timeout:
            key = make_key('memorious', 'timeout', stage)
            if not conn.get(key):
                conn.srem(stages_on_timeout_key, stage)


def get_worker():
    return MemoriousWorker(conn=conn)



import logging
import math

from servicelayer.jobs import Job
from servicelayer.cache import make_key

from memorious.core import conn
from memorious.settings import MAX_QUEUE_LENGTH
from memorious.exc import QueueTooBigError

log = logging.getLogger(__name__)


class Queue(object):
    """Manage the execution of tasks in the system."""

    @classmethod
    def queue(cls, stage, state, data):
        crawler = state.get('crawler')
        job = Job(conn, str(crawler), state['run_id'])
        job_stage = job.get_stage(stage.namespaced_name)
        job_stage.sync()
        queue_length = job_stage.get_status().get('pending')
        if queue_length > MAX_QUEUE_LENGTH:
            msg = "queue for %s:%s too big."
            raise QueueTooBigError(msg % (str(crawler), stage))
        job_stage.queue(payload=data, context=state)

    @classmethod
    def timeout(cls, stage, rate_limit):
        stages_on_timeout = make_key('memorious', 'timeout_stages')
        conn.sadd(stages_on_timeout, stage.namespaced_name)
        stage_timeout_key = make_key(
            'memorious', 'timeout', stage.namespaced_name
        )
        expiry = (rate_limit.interval * rate_limit.unit) / rate_limit.limit
        conn.set(stage_timeout_key, 'true', ex=math.ceil(expiry))
        # Delay the current task without further adding to call count
        rate_limit.comply(amount=0)
